/*         ______   ___    ___ 
 *        /\  _  \ /\_ \  /\_ \ 
 *        \ \ \L\ \\//\ \ \//\ \      __     __   _ __   ___ 
 *         \ \  __ \ \ \ \  \ \ \   /'__`\ /'_ `\/\`'__\/ __`\
 *          \ \ \/\ \ \_\ \_ \_\ \_/\  __//\ \L\ \ \ \//\ \L\ \
 *           \ \_\ \_\/\____\/\____\ \____\ \____ \ \_\\ \____/
 *            \/_/\/_/\/____/\/____/\/____/\/___L\ \/_/ \/___/
 *                                           /\____/
 *                                           \_/__/
 *
 *      DirectDraw windowed gfx driver.
 *
 *      By Isaac Cruz.
 *
 *      See readme.txt for copyright information.
 */

#include "wddraw.h"


static void update_window (void);
static void wddwin_switch_out(void);
static void wddwin_switch_in(void);
static void setup_retrace_proc (void);
static int wnd_set_windowed_coop(void);
static int verify_color_depth (int color_depth);
static int gfx_directx_show_video_bitmap_win(struct BITMAP *bitmap);
static struct BITMAP *init_directx_win(int w, int h, int v_w, int v_h, int color_depth);
static void gfx_directx_win_exit(struct BITMAP *b);


GFX_DRIVER gfx_directx_win =
{
   GFX_DIRECTX_WIN,
   empty_string,
   empty_string,
   "DirectDraw window",
   init_directx_win,
   gfx_directx_win_exit,
   NULL,			// AL_METHOD(int, scroll, (int x, int y));   
   gfx_directx_sync,
   gfx_directx_set_palette,
   NULL,			// AL_METHOD(int, request_scroll, (int x, int y));
   NULL,			// gfx_directx_poll_scroll,
   NULL,			// AL_METHOD(void, enable_triple_buffer, (void));
   gfx_directx_create_video_bitmap,
   gfx_directx_destroy_video_bitmap,
   gfx_directx_show_video_bitmap_win,
   gfx_directx_show_video_bitmap_win,
   gfx_directx_create_system_bitmap,
   gfx_directx_destroy_system_bitmap,
   NULL,			// AL_METHOD(int, set_mouse_sprite, (struct BITMAP *sprite, int xfocus, int yfocus));
   NULL,			// AL_METHOD(int, show_mouse, (struct BITMAP *bmp, int x, int y));
   NULL,			// AL_METHOD(void, hide_mouse, (void));
   NULL,			// AL_METHOD(void, move_mouse, (int x, int y));
   NULL,			// AL_METHOD(void, drawing_mode, (void));
   NULL,                        // AL_METHOD(void, save_video_state, (void*));
   NULL,                        // AL_METHOD(void, restore_video_state, (void*));
   0, 0,			// int w, h;                     /* physical (not virtual!) screen size */
   TRUE,			// int linear;                   /* true if video memory is linear */
   0,				// long bank_size;               /* bank size, in bytes */
   0,				// long bank_gran;               /* bank granularity, in bytes */
   0,				// long vid_mem;                 /* video memory size, in bytes */
   0,				// long vid_phys_base;           /* physical address of video memory */
};

/* offscreen surface that will be blitted to the window:*/
LPDIRECTDRAWSURFACE dd_offscreen = NULL;
LPDIRECTDRAWSURFACE pseudo_screen = NULL;  /* for page-flipping */
BOOL app_active = TRUE;
RECT window_rect;


/* handle_window_size_win:
 *  updates window_rect if window is moved or resized
 */
void handle_window_size_win(void)
{
   /* update window_rect */
   GetClientRect(allegro_wnd, &window_rect);
   ClientToScreen(allegro_wnd, (LPPOINT)&window_rect);
   ClientToScreen(allegro_wnd, (LPPOINT)&window_rect + 1);
   
}


/* update_window
 * function synced with the vertical refresh
 */
static void update_window (void)
{

   /* blit offscreen backbuffer to the window */
   _enter_gfx_critical();
   IDirectDrawSurface_Blt(dd_prim_surface,
                          &window_rect,
                          pseudo_screen, NULL,
                          0, NULL);
   _exit_gfx_critical();   
}


/* wddwin_switch_out:
 *  handle window switched out
 */
static void wddwin_switch_out(void)
{
   if (app_active) {
      app_active = FALSE;
      remove_int(update_window);
   }
}



/* wddwin_switch_in:
 *  handle window switched in
 */
static void wddwin_switch_in(void)
{
   handle_window_size_win();
   if (!app_active) {
      app_active = TRUE;
      setup_retrace_proc();
   }
}


/* setup_retrace_proc
 */
static void setup_retrace_proc (void)
{	
   HRESULT hr;
   int freq;

   /* set correct sync timer speed */
   hr = IDirectDraw_GetMonitorFrequency(directdraw, &freq);
   if ((FAILED(hr)) || (freq < 40) || (freq > 200))
      install_int_ex (update_window, BPS_TO_TIMER(70));
   else
      install_int_ex (update_window, BPS_TO_TIMER(freq));

}


/* wnd_set_windowed_coop
 */
static int wnd_set_windowed_coop(void)
{
   HRESULT hr;
   
   hr = IDirectDraw_SetCooperativeLevel(directdraw, allegro_wnd, DDSCL_NORMAL);
   if (FAILED(hr)) {
      TRACE("SetCooperative level = %s (%x), hwnd = %x\n", win_err_str(hr), hr, allegro_wnd);
      return -1;
   }

   return 0;
}

/* verify_color_depth:
 * compares the color depth requested with the real color depth
 */
static int verify_color_depth (int color_depth)
{
   DDSURFACEDESC surf_desc;
   HRESULT hr;
   int real_depth;
   
   /* get current video mode */
   surf_desc.dwSize = sizeof(surf_desc);
   hr = IDirectDraw_GetDisplayMode(directdraw, &surf_desc);
   if (FAILED(hr)) {
      TRACE("Can't get color format.\n");
      return -1;
   }
   real_depth = surf_desc.ddpfPixelFormat.dwRGBBitCount;
   if (real_depth == color_depth)
      return 0;
   if ((real_depth == 16) && (color_depth == 15))
      return 0;
   return -1; 
}


/* gfx_directx_show_video_bitmap_win:
 */
static int gfx_directx_show_video_bitmap_win(struct BITMAP *bitmap)
{
   if (BMP_EXTRA(bitmap)->surf) {
      pseudo_screen = BMP_EXTRA(bitmap)->surf;
      return 0;
   }
   return -1;
}


/* init_directx_win:
 */
static struct BITMAP *init_directx_win(int w, int h, int v_w, int v_h, int color_depth)
{
   RECT win_size;
   HRESULT hr;
   int id;

   /* Flipping is impossible in windowed mode */
   if ((v_w != w && v_w != 0) || (v_h != h && v_h != 0))
      return NULL;
      
   _enter_critical();

   /* init DirectX */
   if (init_directx() != 0)
      goto Error;
   if (verify_color_depth(color_depth))
      goto Error;
   if (wnd_call_proc(wnd_set_windowed_coop) != 0)
      goto Error;
   if (finalize_directx_init() != 0)
      goto Error;

   /* adjust window */
   wnd_paint_back = TRUE;
   win_size.left = wnd_x = 50;
   win_size.right = 50 + w;
   win_size.top = wnd_y = 50;
   win_size.bottom = 50 + h;
   wnd_width = w;
   wnd_height = h;
   window_rect = win_size;

   AdjustWindowRect(&win_size, GetWindowLong(allegro_wnd, GWL_STYLE), FALSE);
   MoveWindow(allegro_wnd, win_size.left, win_size.top,
   win_size.right - win_size.left, win_size.bottom - win_size.top, TRUE);

   /* create primary surface */
   if (create_primary(w, h, color_depth) != 0)
      goto Error;
      
   /* create offscreen backbuffer */
   id = BMP_ID_VIDEO;
   dd_offscreen = gfx_directx_create_surface(w, h, color_depth, 1, 0, 0);
   if (dd_offscreen == NULL) {
      /* if there's not enough video RAM, try in system RAM */
      id = BMP_ID_SYSTEM;
      dd_offscreen = gfx_directx_create_surface(w, h, color_depth, 0, 0, 0);
      if (dd_offscreen == NULL)
         goto Error;
   }
   
   /* create clipper */
   if (create_clipper(allegro_wnd) != 0)
      goto Error;
   hr = IDirectDrawSurface_SetClipper(dd_prim_surface, dd_clipper);
   if (FAILED(hr))
      goto Error;
   
   if (color_depth == 8) {
      if (create_palette(dd_prim_surface) != 0)
         goto Error;
   }
   else
       gfx_directx_update_color_format(color_depth);

   /* setup Allegro gfx driver */
   if (setup_driver(&gfx_directx_win, w, h, color_depth) != 0)
      goto Error;
   dd_frontbuffer = make_directx_bitmap(dd_offscreen, w, h, color_depth, id);
   
   enable_acceleration(&gfx_directx_win);
   
   set_display_switch_callback(SWITCH_OUT, wddwin_switch_out);
   set_display_switch_callback(SWITCH_IN, wddwin_switch_in);

   setup_retrace_proc();

   _exit_critical();
   
   pseudo_screen = dd_offscreen;
   return dd_frontbuffer;

 Error:
   _exit_critical();

   /* release the DirectDraw object */
   gfx_directx_exit(NULL);

   return NULL;
}

/* gfx_directx_exit:
 */
static void gfx_directx_win_exit(struct BITMAP *b)
{ 
   remove_display_switch_callback(wddwin_switch_out);
   remove_display_switch_callback(wddwin_switch_in);
   
   if (app_active) {
      app_active = FALSE;
      remove_int(update_window);
   }

   if (b)
      clear (b);
   
   /* destroy the offscreen backbuffer used in windowed mode */
   if (dd_offscreen) {
      IDirectDrawSurface_Release(dd_offscreen);
      dd_offscreen = NULL;
   }
      
   /* unlink surface from bitmap */
   if (b)
      b->extra = NULL;

   gfx_directx_exit(NULL);
}